from abc import ABC, abstractmethod
from typing import List, Tuple, Dict, Any, Optional, Union
from pathlib import Path

class Generator(ABC):
    """Base class for generating synthetic data pairs."""
    
    @abstractmethod
    def generate_pairs(
        self, 
        instructions: Optional[List[str]] = None,
        production_logs_path: Optional[Union[str, Path]] = None,
        instruction_col: str = "instruction",
        bad_answer_col: str = "bad_answer",
        **kwargs
    ) -> List[Dict]:
        """Generate pairs of responses (good, bad) for instructions.
        
        Args:
            instructions: Optional list of instructions for synthetic generation
            production_logs_path: Optional path to CSV with real bad answers
            instruction_col: Column name for instructions in CSV
            bad_answer_col: Column name for bad answers in CSV
            **kwargs: Additional arguments for specific implementations
            
        Returns:
            List of dicts containing instruction, good_answer, bad_answer, and source
        """
        pass

class Judge(ABC):
    """Base class for evaluating response pairs."""
    
    @abstractmethod
    def judge_pairs(self,
                    instruction: str,
                    pairs: List[Tuple[str, str]],
                    n_samples: int = 5) -> List[str]:
        """Judge which response in each pair is better.
        Returns: List of 'A' or 'B' indicating which response was better."""
        pass

class Pipeline:
    """Coordinates the generation and evaluation flow."""
    
    def __init__(self,
                 generator: Generator,
                 judge: Judge):
        self.generator = generator
        self.judge = judge
    
    def run(self,
            instructions: Optional[List[str]] = None,
            production_logs_path: Optional[Union[str, Path]] = None,
            n_judge_samples: int = 5,
            **kwargs) -> Dict[str, Any]:
        """Run the full pipeline for instructions.
        
        Args:
            instructions: Optional list of instructions for synthetic generation
            production_logs_path: Optional path to CSV with real bad answers
            n_judge_samples: Number of samples for judge
            **kwargs: Additional arguments passed to generator
            
        Returns:
            Dict containing generated pairs and judgments
        """
        # Generate pairs
        pairs = self.generator.generate_pairs(
            instructions=instructions,
            production_logs_path=production_logs_path,
            **kwargs
        )
        
        # Judge the pairs
        all_judgments = []
        for pair in pairs:
            judgments = self.judge.judge_pairs(
                pair["instruction"], 
                [(pair["good_answer"], pair["bad_answer"])],
                n_judge_samples
            )
            all_judgments.extend(judgments)
        
        return {
            "pairs": pairs,
            "judgments": all_judgments
        } 