from typing import List, Dict, Optional, Union, Callable, Literal
import csv
from pathlib import Path
import pandas as pd
from phinity.core.interfaces import Generator
from phinity.utils.config import Config

# Type aliases for clarity
PromptTemplate = Union[str, Callable[[str], str]]
DataSource = Literal["synthetic", "production_logs_bad", "production_logs_good"]

class LLMGenerator(Generator):
    """
    Flexible generator supporting multiple workflows, always using paper's methodology 
    for generating bad answers (via noisy instructions):
    
    1. Paper methodology: 
       instruction -> good answer
                  -> noisy instruction -> bad answer
                  
    2. From bad answers: 
       (instruction, bad) -> generate good answer
       
    3. From good answers: 
       (instruction, good) -> noisy instruction -> bad answer
    """

    # Default prompt templates
    DEFAULT_GOOD_TEMPLATE = """Given this instruction:
{instruction}

Provide a high-quality response that follows the instruction perfectly."""

    DEFAULT_NOISY_INSTRUCTION_TEMPLATE = """Original instruction:
{instruction}

Generate a modified instruction that is highly relevant but not semantically identical to the instruction above.
The modified instruction should be related but would lead to a different kind of response.

Modified instruction:"""

    def __init__(
        self, 
        good_prompt_template: Optional[PromptTemplate] = None,
        noisy_instruction_template: Optional[PromptTemplate] = None,
        model: str = "gpt-4o-mini"
    ):
        """
        Args:
            good_prompt_template: Template for generating good answers
            noisy_instruction_template: Template for generating noisy instructions
            model: OpenAI model to use
        """
        self.good_prompt_template = good_prompt_template
        self.noisy_instruction_template = noisy_instruction_template
        self.model = model

    def _format_prompt(self, template: Optional[PromptTemplate], default_template: str, **kwargs) -> str:
        """Format a prompt using either custom or default template."""
        try:
            if template is None:
                return default_template.format(**kwargs)
            
            if isinstance(template, str):
                return template.format(**kwargs)
            
            if callable(template):
                return template(**kwargs)
            
            raise ValueError(f"Invalid template type: {type(template)}")
            
        except Exception as e:
            print(f"Error formatting prompt: {str(e)}")
            print(f"Falling back to default template")
            return default_template.format(**kwargs)

    def generate_good_answer(self, instruction: str, current_answer: Optional[str] = None) -> str:
        """Generate a high-quality answer for the instruction."""
        prompt = self._format_prompt(
            self.good_prompt_template,
            self.DEFAULT_GOOD_TEMPLATE,
            instruction=instruction,
            current_answer=current_answer if current_answer else ""
        )
        
        client = Config.get_instance().openai_client
        if client is None:
            raise RuntimeError("OpenAI client not configured. Call Config.setup_openai() first")
            
        response = client.chat.completions.create(
            model=self.model,
            messages=[{"role": "user", "content": prompt}],
            temperature=0.7
        )
        
        return response.choices[0].message.content.strip()

    def generate_noisy_instruction(self, instruction: str) -> str:
        """Generate a modified version of the instruction that's related but different."""
        prompt = self._format_prompt(
            self.noisy_instruction_template,
            self.DEFAULT_NOISY_INSTRUCTION_TEMPLATE,
            instruction=instruction
        )
        
        client = Config.get_instance().openai_client
        if client is None:
            raise RuntimeError("OpenAI client not configured. Call Config.setup_openai() first")
            
        response = client.chat.completions.create(
            model=self.model,
            messages=[{"role": "user", "content": prompt}],
            temperature=0.9
        )
        
        return response.choices[0].message.content.strip()

    def generate_bad_answer_via_noisy(self, instruction: str) -> tuple[str, str]:
        """
        Generate a bad answer using the paper's methodology:
        1. Generate noisy version of instruction
        2. Generate good answer for noisy instruction (becomes bad answer for original)
        
        Returns:
            Tuple of (noisy_instruction, bad_answer)
        """
        noisy_instruction = self.generate_noisy_instruction(instruction)
        bad_answer = self.generate_good_answer(noisy_instruction)
        return noisy_instruction, bad_answer

    def generate_pairs(
        self, 
        instructions: Optional[List[str]] = None,
        logs_path: Optional[Union[str, Path]] = None,
        logs_type: Optional[DataSource] = None,
        instruction_col: str = "instruction",
        answer_col: Optional[str] = None,
        **kwargs
    ) -> List[Dict]:
        """
        Generate instruction-response pairs using various methods.
        Always uses paper's methodology for generating bad answers.
        """
        pairs = []
        
        # Method 1: Paper's synthetic generation
        if instructions:
            for instr in instructions:
                good_answer = self.generate_good_answer(instr)
                noisy_instruction, bad_answer = self.generate_bad_answer_via_noisy(instr)
                
                pairs.append({
                    "instruction": instr,
                    "good_answer": good_answer,
                    "bad_answer": bad_answer,
                    "noisy_instruction": noisy_instruction,
                    "source": "synthetic"
                })
        
        # Method 2 & 3: Production logs
        if logs_path:
            if not logs_type:
                raise ValueError("logs_type must be specified when using logs_path")
            if not answer_col:
                raise ValueError("answer_col must be specified when using logs_path")
                
            df = pd.read_csv(logs_path)
            required_cols = [instruction_col, answer_col]
            missing_cols = [col for col in required_cols if col not in df.columns]
            if missing_cols:
                raise ValueError(f"Missing required columns: {missing_cols}")
            
            for _, row in df.iterrows():
                instr = row[instruction_col]
                current_answer = row[answer_col]
                
                if logs_type == "production_logs_bad":
                    # Pass both instruction and current answer for context
                    good_answer = self.generate_good_answer(instr, current_answer)
                    pairs.append({
                        "instruction": instr,
                        "good_answer": good_answer,
                        "bad_answer": current_answer,
                        "source": logs_type
                    })
                    
                elif logs_type == "production_logs_good":
                    noisy_instruction, bad_answer = self.generate_bad_answer_via_noisy(instr)
                    pairs.append({
                        "instruction": instr,
                        "good_answer": current_answer,
                        "bad_answer": bad_answer,
                        "noisy_instruction": noisy_instruction,
                        "source": logs_type
                    })
                    
        if not pairs:
            raise ValueError(
                "Must provide either instructions for synthetic generation "
                "or logs_path with appropriate logs_type"
            )
                
        return pairs 